/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// RSP Processor stuff
#include "stdafx.h"

#include <shlwapi.h>

#include "CPU.h"
#include "RSP.h"
#include "RDP.h"
#include "Memory.h"
#include "Debug.h"
#include "Dump.h"			// For Dump_GetDumpDirectory()
#include "OpCode.h"
#include "Registers.h"
#include "ultra_rcp.h"
#include "DBGConsole.h"

#include "MainWindow.h"
#include "R4300.h"

#include "ultra_r4300.h"

// We need similar registers to the main CPU

// 32 32-bit General Purpose Registers
DWORD	g_dwRSPGPR[32];

// 32 8*16bit Vector registers
static WORD	g_wVR[32][8];
static WORD	g_wTemp[2][8];		// Temporary registers for vector calculations

static DWORD g_dwCarryFlags = 0;

static DWORD g_dwRSPOp = 0;
static DWORD g_nRSPDelay = NO_DELAY;
static DWORD g_dwNewRSPPC = 0;

BOOL	g_bRSPHalted		= TRUE;

// We also have some vector registers here


typedef void (*RSPInstruction)();


static void RSP_Unk();
static void RSP_Special();
static void RSP_RegImm();
static void RSP_J();
static void RSP_JAL();
static void RSP_BEQ();
static void RSP_BNE();
static void RSP_BLEZ();
static void RSP_BGTZ();
static void RSP_ADDI();
static void RSP_ADDIU();
static void RSP_SLTI();
static void RSP_SLTIU();
static void RSP_ANDI();
static void RSP_ORI();
static void RSP_XORI();
static void RSP_LUI();
static void RSP_CoPro0();
static void RSP_CoPro2();
static void RSP_LB();
static void RSP_LH();
static void RSP_LW();
static void RSP_LBU();
static void RSP_LHU();
static void RSP_SB();
static void RSP_SH();
static void RSP_SW();
static void RSP_LWC2();
static void RSP_SWC2();


static void RSP_Special_Unk();
static void RSP_Special_SLL();
static void RSP_Special_SRL();
static void RSP_Special_SRA();
static void RSP_Special_SLLV();
static void RSP_Special_SRLV();
static void RSP_Special_SRAV();
static void RSP_Special_JR();
static void RSP_Special_JALR();
static void RSP_Special_BREAK();
static void RSP_Special_ADD();
static void RSP_Special_ADDU();
static void RSP_Special_SUB();
static void RSP_Special_SUBU();
static void RSP_Special_AND();
static void RSP_Special_OR();
static void RSP_Special_XOR();
static void RSP_Special_NOR();
static void RSP_Special_SLT();
static void RSP_Special_SLTU();

static void RSP_RegImm_Unk();
static void RSP_RegImm_BLTZ();
static void RSP_RegImm_BGEZ();
static void RSP_RegImm_BLTZAL();
static void RSP_RegImm_BGEZAL();


static void RSP_Cop0_Unk();
static void RSP_Cop0_MFC0();
static void RSP_Cop0_MTC0();

static void RSP_LWC2_Unk();
static void RSP_LWC2_LBV();
static void RSP_LWC2_LSV();
static void RSP_LWC2_LLV();
static void RSP_LWC2_LDV();
static void RSP_LWC2_LQV();
static void RSP_LWC2_LRV();
static void RSP_LWC2_LPV();
static void RSP_LWC2_LUV();
static void RSP_LWC2_LHV();
static void RSP_LWC2_LFV();
static void RSP_LWC2_LWV();
static void RSP_LWC2_LTV();

static void RSP_SWC2_Unk();
static void RSP_SWC2_SBV();
static void RSP_SWC2_SSV();
static void RSP_SWC2_SLV();
static void RSP_SWC2_SDV();
static void RSP_SWC2_SQV();
static void RSP_SWC2_SRV();
static void RSP_SWC2_SPV();
static void RSP_SWC2_SUV();
static void RSP_SWC2_SHV();
static void RSP_SWC2_SFV();
static void RSP_SWC2_SWV();
static void RSP_SWC2_STV();

static void RSP_Cop2_Unk();
static void RSP_Cop2_MFC2();
static void RSP_Cop2_CFC2();
static void RSP_Cop2_MTC2();
static void RSP_Cop2_CTC2();
static void RSP_Cop2_Vector();

static void RSP_Cop2V_Unk();
static void RSP_Cop2V_VMULF();
static void RSP_Cop2V_VMULU();
static void RSP_Cop2V_VRNDP();
static void RSP_Cop2V_VMULQ();
static void RSP_Cop2V_VMUDL();
static void RSP_Cop2V_VMUDM();
static void RSP_Cop2V_VMUDN();
static void RSP_Cop2V_VMUDH();
static void RSP_Cop2V_VMACF();
static void RSP_Cop2V_VMACU();
static void RSP_Cop2V_VRNDN();
static void RSP_Cop2V_VMACQ();
static void RSP_Cop2V_VMADL();
static void RSP_Cop2V_VMADM();
static void RSP_Cop2V_VMADN();
static void RSP_Cop2V_VMADH();
static void RSP_Cop2V_VADD();
static void RSP_Cop2V_VSUB();
static void RSP_Cop2V_VSUT();
static void RSP_Cop2V_VABS();
static void RSP_Cop2V_VADDC();
static void RSP_Cop2V_VSUBC();
static void RSP_Cop2V_VADDB();
static void RSP_Cop2V_VSUBB();
static void RSP_Cop2V_VACCB();
static void RSP_Cop2V_VSUCB();
static void RSP_Cop2V_VSAD();
static void RSP_Cop2V_VSAC();
static void RSP_Cop2V_VSUM();
static void RSP_Cop2V_VSAW();
static void RSP_Cop2V_VLT();
static void RSP_Cop2V_VEQ();
static void RSP_Cop2V_VNE();
static void RSP_Cop2V_VGE();
static void RSP_Cop2V_VCL();
static void RSP_Cop2V_VCH();
static void RSP_Cop2V_VCR();
static void RSP_Cop2V_VMRG();
static void RSP_Cop2V_VAND();
static void RSP_Cop2V_VNAND();
static void RSP_Cop2V_VOR();
static void RSP_Cop2V_VNOR();
static void RSP_Cop2V_VXOR();
static void RSP_Cop2V_VNXOR();
static void RSP_Cop2V_VRCP();
static void RSP_Cop2V_VRCPL();
static void RSP_Cop2V_VRCPH();
static void RSP_Cop2V_VMOV();
static void RSP_Cop2V_VRSQ();
static void RSP_Cop2V_VRSQL();
static void RSP_Cop2V_VRSQH();



/*
    CPU: Instructions encoded by opcode field.
    31---------26---------------------------------------------------0
    |  opcode   |                                                   |
    ------6----------------------------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 | *1    | *2    |   J   |  JAL  |  BEQ  |  BNE  | BLEZ  | BGTZ  |
001 | ADDI  | ADDIU | SLTI  | SLTIU | ANDI  |  ORI  | XORI  |  LUI  |
010 | *3    |  ---  |  *4   |  ---  |  ---  |  ---  |  ---  |  ---  |
011 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
100 |  LB   |  LH   |  ---  |  LW   |  LBU  |  LHU  |  ---  |  ---  |
101 |  SB   |  SH   |  ---  |  SW   |  ---  |  ---  |  ---  |  ---  |
110 |  ---  |  ---  | *LWC2 |  ---  |  ---  |  ---  |  ---  |  ---  |
111 |  ---  |  ---  | *SWC2 |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
     *1 = SPECIAL, see SPECIAL list    *2 = REGIMM, see REGIMM list
     *3 = COP0                         *4 = COP2
     *LWC2 = RSP Load instructions     *SWC2 = RSP Store instructions
*/


// Opcode Jump Table
RSPInstruction RSP_Instruction[64] = {
	RSP_Special, RSP_RegImm, RSP_J, RSP_JAL, RSP_BEQ, RSP_BNE, RSP_BLEZ, RSP_BGTZ,
	RSP_ADDI, RSP_ADDIU, RSP_SLTI, RSP_SLTIU, RSP_ANDI, RSP_ORI, RSP_XORI, RSP_LUI,
	RSP_CoPro0, RSP_Unk, RSP_CoPro2, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk,
	RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk,
	RSP_LB, RSP_LH, RSP_Unk, RSP_LW, RSP_LBU, RSP_LHU, RSP_Unk, RSP_Unk,
	RSP_SB, RSP_SH, RSP_Unk, RSP_SW, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk,
	RSP_Unk, RSP_Unk, RSP_LWC2, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk,
	RSP_Unk, RSP_Unk, RSP_SWC2, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk, RSP_Unk
};


/*
    RSP SPECIAL: Instr. encoded by function field when opcode field = SPECIAL.
    31---------26-----------------------------------------5---------0
    | = SPECIAL |                                         | function|
    ------6----------------------------------------------------6-----
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 |  SLL  |  ---  |  SRL  |  SRA  | SLLV  |  ---  | SRLV  | SRAV  |
001 |  JR   |  JALR |  ---  |  ---  |  ---  | BREAK |  ---  |  ---  |
010 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
011 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
100 |  ADD  | ADDU  |  SUB  | SUBU  |  AND  |  OR   |  XOR  |  NOR  |
101 |  ---  |  ---  |  SLT  | SLTU  |  ---  |  ---  |  ---  |  ---  |
110 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
111 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
 */

// SpecialOpCode Jump Table
RSPInstruction RSPSpecialInstruction[64] = {
	RSP_Special_SLL, RSP_Special_Unk,  RSP_Special_SRL, RSP_Special_SRA,  RSP_Special_SLLV, RSP_Special_Unk,   RSP_Special_SRLV, RSP_Special_SRAV,
	RSP_Special_JR,  RSP_Special_JALR, RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk,  RSP_Special_BREAK, RSP_Special_Unk,  RSP_Special_Unk,
	RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk,  RSP_Special_Unk,   RSP_Special_Unk,  RSP_Special_Unk,
	RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk,  RSP_Special_Unk,   RSP_Special_Unk,  RSP_Special_Unk,
	RSP_Special_ADD, RSP_Special_ADDU, RSP_Special_SUB, RSP_Special_SUBU, RSP_Special_AND,  RSP_Special_OR,    RSP_Special_XOR,  RSP_Special_NOR,
	RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_SLT, RSP_Special_SLTU, RSP_Special_Unk,  RSP_Special_Unk,   RSP_Special_Unk,  RSP_Special_Unk,
	RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk,  RSP_Special_Unk,   RSP_Special_Unk,  RSP_Special_Unk,
	RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk, RSP_Special_Unk,  RSP_Special_Unk,  RSP_Special_Unk,   RSP_Special_Unk,  RSP_Special_Unk
};


/*
    REGIMM: Instructions encoded by the rt field when opcode field = REGIMM.
    31---------26----------20-------16------------------------------0
    | = REGIMM  |          |   rt    |                              |
    ------6---------------------5------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
 00 | BLTZ  | BGEZ  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 01 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 10 |BLTZAL |BGEZAL |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 11 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
 */

RSPInstruction RSPRegImmInstruction[32] = {

	RSP_RegImm_BLTZ,   RSP_RegImm_BGEZ,   RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk,
	RSP_RegImm_Unk,    RSP_RegImm_Unk,    RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk,
	RSP_RegImm_BLTZAL, RSP_RegImm_BGEZAL, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk,
	RSP_RegImm_Unk,    RSP_RegImm_Unk,    RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk, RSP_RegImm_Unk
};



/*
    COP0: Instructions encoded by the fmt field when opcode = COP0.
    31--------26-25------21 ----------------------------------------0
    |  = COP0   |   fmt   |                                         |
    ------6----------5-----------------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
 00 | MFC0  |  ---  |  ---  |  ---  | MTC0  |  ---  |  ---  |  ---  |
 01 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 10 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 11 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/

// COP0 Jump Table
RSPInstruction RSPCop0Instruction[32] = {
	RSP_Cop0_MFC0, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_MTC0, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk,
	RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk,
	RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk,
	RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk, RSP_Cop0_Unk,
};



/*
    RSP Load: Instr. encoded by rd field when opcode field = LWC2
    31---------26-------------------15-------11---------------------0
    |  110010   |                   |   rd   |                      |
    ------6-----------------------------5----------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
 00 |  LBV  |  LSV  |  LLV  |  LDV  |  LQV  |  LRV  |  LPV  |  LUV  |
 01 |  LHV  |  LFV  |  LWV  |  LTV  |  ---  |  ---  |  ---  |  ---  |
 10 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 11 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/

// LWC2 Jump Table
RSPInstruction RSPLWC2Instruction[32] = {
	RSP_LWC2_LBV, RSP_LWC2_LSV, RSP_LWC2_LLV, RSP_LWC2_LDV, RSP_LWC2_LQV, RSP_LWC2_LRV, RSP_LWC2_LPV, RSP_LWC2_LUV,
	RSP_LWC2_LHV, RSP_LWC2_LFV, RSP_LWC2_LWV, RSP_LWC2_LTV, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk,
	RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk,
	RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk, RSP_LWC2_Unk,
};

/*
    RSP Store: Instr. encoded by rd field when opcode field = SWC2
    31---------26-------------------15-------11---------------------0
    |  111010   |                   |   rd   |                      |
    ------6-----------------------------5----------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
 00 |  SBV  |  SSV  |  SLV  |  SDV  |  SQV  |  SRV  |  SPV  |  SUV  |
 01 |  SHV  |  SFV  |  SWV  |  STV  |  ---  |  ---  |  ---  |  ---  |
 10 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 11 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/

// SWC2 Jump Table
RSPInstruction RSPSWC2Instruction[32] = {
	RSP_SWC2_SBV, RSP_SWC2_SSV, RSP_SWC2_SLV, RSP_SWC2_SDV, RSP_SWC2_SQV, RSP_SWC2_SRV, RSP_SWC2_SPV, RSP_SWC2_SUV,
	RSP_SWC2_SHV, RSP_SWC2_SFV, RSP_SWC2_SWV, RSP_SWC2_STV, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk,
	RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk,
	RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk, RSP_SWC2_Unk,
};


/*
    COP2: Instructions encoded by the fmt field when opcode = COP2.
    31--------26-25------21 ----------------------------------------0
    |  = COP2   |   fmt   |                                         |
    ------6----------5-----------------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
 00 | MFC2  |  ---  | CFC2  |  ---  | MTC2  |  ---  | CTC2  |  ---  |
 01 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 10 |  *1   |  *1   |  *1   |  *1   |  *1   |  *1   |  *1   |  *1   |
 11 |  *1   |  *1   |  *1   |  *1   |  *1   |  *1   |  *1   |  *1   |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
     *1 = Vector opcode
*/

RSPInstruction RSPCop2Instruction[32] = {
	RSP_Cop2_MFC2, RSP_Cop2_Unk, RSP_Cop2_CFC2, RSP_Cop2_Unk, RSP_Cop2_MTC2, RSP_Cop2_Unk, RSP_Cop2_CTC2, RSP_Cop2_Unk,
	RSP_Cop2_Unk, RSP_Cop2_Unk, RSP_Cop2_Unk, RSP_Cop2_Unk, RSP_Cop2_Unk, RSP_Cop2_Unk, RSP_Cop2_Unk, RSP_Cop2_Unk, 
	RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector,
	RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector, RSP_Cop2_Vector
};

/*
    Vector opcodes: Instr. encoded by the function field when opcode = COP2.
    31---------26---25------------------------------------5---------0
    |  = COP2   | 1 |                                     | function|
    ------6-------1--------------------------------------------6-----
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 | VMULF | VMULU | VRNDP | VMULQ | VMUDL | VMUDM | VMUDN | VMUDH |
001 | VMACF | VMACU | VRNDN | VMACQ | VMADL | VMADM | VMADN | VMADH |
010 | VADD  | VSUB  | VSUT? | VABS  | VADDC | VSUBC | VADDB?| VSUBB?|
011 | VACCB?| VSUCB?| VSAD? | VSAC? | VSUM? | VSAW  |  ---  |  ---  |
100 |  VLT  |  VEQ  |  VNE  |  VGE  |  VCL  |  VCH  |  VCR  | VMRG  |
101 | VAND  | VNAND |  VOR  | VNOR  | VXOR  | VNXOR |  ---  |  ---  |
110 | VRCP  | VRCPL | VRCPH | VMOV  | VRSQ  | VRSQL | VRSQH |  ---  |
110 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
    Comment: Those with a ? in the end of them may not exist
*/
RSPInstruction RSPCop2VInstruction[64] = {
	RSP_Cop2V_VMULF, RSP_Cop2V_VMULU, RSP_Cop2V_VRNDP, RSP_Cop2V_VMULQ, RSP_Cop2V_VMUDL, RSP_Cop2V_VMUDM, RSP_Cop2V_VMUDN, RSP_Cop2V_VMUDH,
	RSP_Cop2V_VMACF, RSP_Cop2V_VMACU, RSP_Cop2V_VRNDN, RSP_Cop2V_VMACQ, RSP_Cop2V_VMADL, RSP_Cop2V_VMADM, RSP_Cop2V_VMADN, RSP_Cop2V_VMADH,
	RSP_Cop2V_VADD,  RSP_Cop2V_VSUB,  RSP_Cop2V_VSUT,  RSP_Cop2V_VABS,  RSP_Cop2V_VADDC, RSP_Cop2V_VSUBC, RSP_Cop2V_VADDB, RSP_Cop2V_VSUBB,
	RSP_Cop2V_VACCB, RSP_Cop2V_VSUCB, RSP_Cop2V_VSAD,  RSP_Cop2V_VSAC,  RSP_Cop2V_VSUM,  RSP_Cop2V_VSAW,  RSP_Cop2V_Unk,   RSP_Cop2V_Unk,
	RSP_Cop2V_VLT,   RSP_Cop2V_VEQ,   RSP_Cop2V_VNE,   RSP_Cop2V_VGE,   RSP_Cop2V_VCL,   RSP_Cop2V_VCH,   RSP_Cop2V_VCR,   RSP_Cop2V_VMRG,
	RSP_Cop2V_VAND,  RSP_Cop2V_VNAND, RSP_Cop2V_VOR,   RSP_Cop2V_VNOR,  RSP_Cop2V_VXOR,  RSP_Cop2V_VNXOR, RSP_Cop2V_Unk,   RSP_Cop2V_Unk,
	RSP_Cop2V_VRCP,  RSP_Cop2V_VRCPL, RSP_Cop2V_VRCPH, RSP_Cop2V_VMOV,  RSP_Cop2V_VRSQ,  RSP_Cop2V_VRSQL, RSP_Cop2V_VRSQH, RSP_Cop2V_Unk,
	RSP_Cop2V_Unk,   RSP_Cop2V_Unk,   RSP_Cop2V_Unk,   RSP_Cop2V_Unk,   RSP_Cop2V_Unk,   RSP_Cop2V_Unk,   RSP_Cop2V_Unk,   RSP_Cop2V_Unk
};

/*
	vadd   $v1, $v2, $v3          0 1 2 3 4 5 6 7
	vadd   $v1, $v2, $v3[0q]      0 1 2 3 0 1 2 3
	vadd   $v1, $v2, $v3[1q]      4 5 6 7 4 5 6 7
	vadd   $v1, $v2, $v3[0h]      0 1 0 1 0 1 0 1
	vadd   $v1, $v2, $v3[1h]      2 3 2 3 2 3 2 3
	vadd   $v1, $v2, $v3[2h]      4 5 4 5 4 5 4 5
	vadd   $v1, $v2, $v3[3h]      6 7 6 7 6 7 6 7 
	vadd   $v1, $v2, $v3[0]       0 0 0 0 0 0 0 0
	vadd   $v1, $v2, $v3[1]       1 1 1 1 1 1 1 1
	vadd   $v1, $v2, $v3[2]       2 2 2 2 2 2 2 2
	vadd   $v1, $v2, $v3[3]       3 3 3 3 3 3 3 3
	vadd   $v1, $v2, $v3[4]       4 4 4 4 4 4 4 4
	vadd   $v1, $v2, $v3[5]       5 5 5 5 5 5 5 5
	vadd   $v1, $v2, $v3[6]       6 6 6 6 6 6 6 6
	vadd   $v1, $v2, $v3[7]       7 7 7 7 7 7 7 7
*/

#define LOAD_VECTORS(dwA, dwB, dwPattern)						\
	for (DWORD i = 0; i < 8; i++) {								\
		g_wTemp[0][i] = g_wVR[dwA][i];							\
		g_wTemp[1][i] = g_wVR[dwB][g_Pattern[dwPattern][i]];	\
	}

DWORD g_Pattern[16][8] = {
	{ 0, 1, 2, 3, 4, 5, 6, 7 },
	{ 0, 1, 2, 3, 4, 5, 6, 7 },	//?
	{ 0, 0, 2, 2, 4, 4, 6, 6 },
	{ 1, 1, 3, 3, 5, 5, 7, 7 },

	{ 0, 0, 0, 0, 4, 4, 4, 4 },
	{ 1, 1, 1, 1, 5, 5, 5, 5 }, 
	{ 2, 2, 2, 2, 6, 6, 6, 6 },
	{ 3, 3, 3, 3, 7, 7, 7, 7 },

	{ 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 1, 1, 1, 1, 1, 1, 1, 1 },
	{ 2, 2, 2, 2, 2, 2, 2, 2 },
	{ 3, 3, 3, 3, 3, 3, 3, 3 },
	{ 4, 4, 4, 4, 4, 4, 4, 4 },
	{ 5, 5, 5, 5, 5, 5, 5, 5 },
	{ 6, 6, 6, 6, 6, 6, 6, 6 },
	{ 7, 7, 7, 7, 7, 7, 7, 7 }
};

void RSP_Reset(void)
{
	for (DWORD dwReg = 0; dwReg < 32; dwReg++)
		g_dwRSPGPR[dwReg] = 0;

	StopRSP();
}


void RSPStepHalted() {}

/*void RSPDump()
{
	//char code[33];
	char opinfo[200];

	DWORD dwPC;
	for (dwPC = 0x04001000; dwPC < 0x04002000; dwPC+=4)
	{
		DWORD dwOpCode = Read32Bits(dwPC);

		SprintRSPOpCodeInfo(opinfo, dwPC, dwOpCode);

		DPF(DEBUG_UCODE, opinfo);
	}

	DWORD dwMem;
	for (dwMem = 0x04000000; dwMem < 0x04001000; dwMem+=2)
	{
		WORD wData = Read16Bits(dwMem);

		wsprintf(opinfo, "0x%08x: 0x%04x", dwMem, wData);

		DPF(DEBUG_UCODE, opinfo);
	}
}*/

void RSPDump(DWORD * pBase, DWORD dwPCBase, DWORD dwLen)
{
	TCHAR opinfo[400];
	TCHAR szFileName[MAX_PATH+1];
	TCHAR szFilePath[MAX_PATH+1];
	FILE * fp;

	
	Dump_GetDumpDirectory(szFilePath, TEXT("rsp_dumps"));
	wsprintf(szFileName, TEXT("rsp_dump.txt"));
	PathAppend(szFilePath, szFileName);


	DBGConsole_Msg(0, "Dumping rsp stuff to %s", szFilePath);
	
	fp = fopen(szFilePath, "w");
	if (fp == NULL)
		return;
	
	DWORD dwIndex;
	for (dwIndex = 0; dwIndex < 0x1000; dwIndex+=4)
	{
		DWORD dwOpCode;
		DWORD pc = dwIndex&0x1FFF;
		dwOpCode = pBase[dwIndex/4];

		fprintf(fp, "0x%08x: 0x%08x\n", pc, dwOpCode);
		//fprintf(fp, "<0x%08x>\n", dwOpCode);
	}

	for (dwIndex = 0x1000; dwIndex < 0x2000; dwIndex+=4)
	{
		DWORD dwOpCode;
		DWORD pc = dwIndex&0x1FFF;
		dwOpCode = pBase[dwIndex/4];

		SprintRSPOpCodeInfo(opinfo, pc, dwOpCode);
		fprintf(fp, "0x%08x: <0x%08x> %s\n", pc, dwOpCode, opinfo);

		//DWORD dwOpCode2 = Read32Bits(PHYS_TO_K1(dwPCBase +(dwIndex-0x1000)));
		//SprintRSPOpCodeInfo(opinfo, pc, dwOpCode2);
		//fprintf(fp, "0x%08x: <0x%08x> %s\n", PHYS_TO_K1(dwPCBase +(dwIndex-0x1000)),dwOpCode2, opinfo);
	}
	
	fclose(fp);
}


void RSPStep(void)
{


	// Fetch next instruction
	DWORD dwPC = RSPPC();
	g_dwRSPOp = *(DWORD *)((BYTE *)MemoryBuffers[MEM_SP_MEM] + ((dwPC&0x0FFF) | 0x1000));

//	RSPDump((DWORD*)((BYTE *)MemoryBuffers[MEM_SP_MEM]),// + 0x1000),
//		0x04001000, 0x1000);

	RSP_Instruction[op(g_dwRSPOp)]();

	switch (g_nRSPDelay) {
		case DO_DELAY:
			// We've got a delayed instruction to execute. Increment
			// PC as normal, so that subsequent instruction is executed
			*(DWORD *)(MemoryBuffers[MEM_SP_PC_REG]) += 4;
			g_nRSPDelay = EXEC_DELAY;
			break;
		case EXEC_DELAY:
			// We've just executed the delayed instr. Now carry out jump
			//  as stored in g_dwNewPC;		
			*(DWORD *)(MemoryBuffers[MEM_SP_PC_REG]) = g_dwNewRSPPC;
			g_nRSPDelay = NO_DELAY;
			break;
		case NO_DELAY:
			// Normal operation - just increment the PC
			*(DWORD *)(MemoryBuffers[MEM_SP_PC_REG]) += 4;
			break;
	}
}





void RSP_Unk()		{ WARN_NOEXIST("RSP_Unk"); }
void RSP_Special()	{ RSPSpecialInstruction[funct(g_dwRSPOp)](); }
void RSP_RegImm()	{ RSPRegImmInstruction[rt(g_dwRSPOp)](); }
void RSP_CoPro0()	{ RSPCop0Instruction[cop0fmt(g_dwRSPOp)](); }		
void RSP_CoPro2()	{ RSPCop2Instruction[rs(g_dwRSPOp)](); }		
void RSP_LWC2()		{ RSPLWC2Instruction[rd(g_dwRSPOp)](); }
void RSP_SWC2()		{ RSPSWC2Instruction[rd(g_dwRSPOp)](); }

void RSP_J() {			// Jump
	
	// Jump
	g_dwNewRSPPC = (RSPPC() & 0xF0000000) | (target(g_dwRSPOp)<<2);
	g_nRSPDelay = DO_DELAY;


}
void RSP_JAL() {				// Jump And Link

	g_dwRSPGPR[31] = (s32)(RSPPC() + 8);		// Store return address

	g_dwNewRSPPC = (RSPPC() & 0xF0000000) | (target(g_dwRSPOp)<<2);
	g_nRSPDelay = DO_DELAY;
}

void RSP_BEQ() {					// Branch on Equal
	//branch if rs == rt
	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);

	if (g_dwRSPGPR[dwRS] == g_dwRSPGPR[dwRT]) {
		s16 wOffset = (s16)imm(g_dwRSPOp);
		g_dwNewRSPPC = RSPPC() + ((s32)wOffset<<2) + 4;
		g_nRSPDelay = DO_DELAY;
	}
}
void RSP_BNE() {				// Branch on Not Equal
	//branch if rs == rt
	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);

	if (g_dwRSPGPR[dwRS] != g_dwRSPGPR[dwRT]) {
		s16 wOffset = (s16)imm(g_dwRSPOp);
		g_dwNewRSPPC = RSPPC() + ((s32)wOffset<<2) + 4;
		g_nRSPDelay = DO_DELAY;
	}

 }
void RSP_BLEZ() {			// Branch on Less than or Equal to Zero

	//branch if rs <= 0
	DWORD dwRS = rs(g_dwRSPOp);

	if ((s32)g_dwRSPGPR[dwRS] <= 0) {
		s16 wOffset = (s16)imm(g_dwRSPOp);
		g_dwNewRSPPC = RSPPC() + (s32)wOffset*4 + 4;
		g_nRSPDelay = DO_DELAY;
	} 
}
void RSP_BGTZ() {			// Branch on Greater Than to Zero

	//branch if rs > 0
	DWORD dwRS = rs(g_dwRSPOp);

	if ((s32)g_dwRSPGPR[dwRS] > 0) {
		s16 wOffset = (s16)imm(g_dwRSPOp);
		g_dwNewRSPPC = RSPPC() + (s32)wOffset*4 + 4;
		g_nRSPDelay = DO_DELAY;
	} 

}
void RSP_ADDI() {			// ADD Immediate
	
	//rt = rs + immediate
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRS = rs(g_dwRSPOp);
	s16 wData = (s16)imm(g_dwRSPOp);

	g_dwRSPGPR[dwRT] = g_dwRSPGPR[dwRS] + (s32)wData;

	// Generates overflow exception
}
void RSP_ADDIU() {				// ADD Immediate Unsigned -- Same as above, but doesn't generate exception
	
	//rt = rs + immediate
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRS = rs(g_dwRSPOp);
	s16 wData = (s16)imm(g_dwRSPOp);

	g_dwRSPGPR[dwRT] = g_dwRSPGPR[dwRS] + (s32)wData;
}

void RSP_SLTI() { WARN_NOIMPL("RSP: SLTI"); }
void RSP_SLTIU() { WARN_NOIMPL("RSP: SLTIU"); }
void RSP_ANDI() {				// AND Immediate
		
	//rt = rs & immediate
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRS = rs(g_dwRSPOp);
	u16 wData = (u16)imm(g_dwRSPOp);

	g_dwRSPGPR[dwRT] = g_dwRSPGPR[dwRS] & (u32)wData;
}
void RSP_ORI() {			// OR Immediate
	
	//rt = rs | immediate
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRS = rs(g_dwRSPOp);
	u16 wData = (u16)imm(g_dwRSPOp);

	g_dwRSPGPR[dwRT] = g_dwRSPGPR[dwRS] | (u32)wData;
}

void RSP_XORI() {						// XOR Immediate
	
	//rt = rs ^ immediate
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRS = rs(g_dwRSPOp);
	u16 wData = (u16)imm(g_dwRSPOp);

	g_dwRSPGPR[dwRT] = g_dwRSPGPR[dwRS] ^ (u32)wData;

}
void RSP_LUI() {
	//dwRS == 0?
	DWORD dwRT = rt(g_dwRSPOp);
	s16 wData = (s16)imm(g_dwRSPOp);

	g_dwRSPGPR[dwRT] = (wData<<16);
	// Keep lower 16bits?
}

void RSP_LB()
{

	DWORD dwBase = rs(g_dwRSPOp);
	s16 wOffset = (s16)imm(g_dwRSPOp);
	DWORD dwRT  = rt(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + (s32)wOffset) & 0x0FFF;

	char nData = Read8Bits(PHYS_TO_K1(0x04000000 + dwAddress));

	g_dwRSPGPR[dwRT] = (s32)(s16)nData;

}
void RSP_LH() {				// Load Halfword

	DWORD dwBase = rs(g_dwRSPOp);
	s16 wOffset = (s16)imm(g_dwRSPOp);
	DWORD dwRT  = rt(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + (s32)wOffset) & 0x0FFF;

	WORD wData = Read16Bits(PHYS_TO_K1(0x04000000 + dwAddress));

	g_dwRSPGPR[dwRT] = (s32)(s16)wData;

}
void RSP_LW() {						// Load Word

	DWORD dwBase = rs(g_dwRSPOp);
	s16 wOffset = (s16)imm(g_dwRSPOp);
	DWORD dwRT  = rt(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + (s32)wOffset) & 0x0FFF;

	DWORD dwData = Read32Bits(PHYS_TO_K1(0x04000000 + dwAddress));

	g_dwRSPGPR[dwRT] = (s32)dwData;

}
void RSP_LBU() { 

	DWORD dwBase = rs(g_dwRSPOp);
	s16 wOffset = (s16)imm(g_dwRSPOp);
	DWORD dwRT  = rt(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + (s32)wOffset) & 0x0FFF;

	BYTE nData = Read8Bits(PHYS_TO_K1(0x04000000 + dwAddress));

	g_dwRSPGPR[dwRT] = (u32)(u16)nData;

}
void RSP_LHU() {

	DWORD dwBase = rs(g_dwRSPOp);
	s16 wOffset = (s16)imm(g_dwRSPOp);
	DWORD dwRT  = rt(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + (s32)wOffset) & 0x0FFF;

	WORD wData = Read16Bits(PHYS_TO_K1(0x04000000 + dwAddress));

	g_dwRSPGPR[dwRT] = (u32)wData;

}
void RSP_SB() {

	DWORD dwBase = rs(g_dwRSPOp);
	s16 wOffset = (s16)imm(g_dwRSPOp);
	DWORD dwRT  = rt(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + (s32)wOffset) & 0x0FFF;

	BYTE nData = (BYTE)g_dwRSPGPR[dwRT];

	Write8Bits(PHYS_TO_K1(0x04000000 + dwAddress), nData);

}

void RSP_SH() {						// Store Halfword

	DWORD dwBase = rs(g_dwRSPOp);
	s16 wOffset = (s16)imm(g_dwRSPOp);
	DWORD dwRT  = rt(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + (s32)wOffset) & 0x0FFF;

	u16 wData =	(u16)g_dwRSPGPR[dwRT];

	Write16Bits(PHYS_TO_K1(0x04000000 + dwAddress), wData);
}

void RSP_SW() { 


	DWORD dwBase = rs(g_dwRSPOp);
	s16 wOffset = (s16)imm(g_dwRSPOp);
	DWORD dwRT  = rt(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + (s32)wOffset) & 0x0FFF;

	DWORD dwData =	g_dwRSPGPR[dwRT];

	Write32Bits(PHYS_TO_K1(0x04000000 + dwAddress), dwData);

}

void RSP_Special_Unk() {
	/*CPUHalt(); WARN_NOEXIST("RSP_Special_Unk");
	DPF(DEBUG_ALWAYS, "RSP: PC: 0x%08x Unknown Op: 0x%08x", RSPPC(), g_dwRSPOp);
	*/
}

void RSP_Special_SLL() {				// Shift word Left Logical

	//DWORD dwRS = rs(g_dwRSPOp);	// Should be == 0
	DWORD dwSA = sa(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	g_dwRSPGPR[dwRD] = g_dwRSPGPR[dwRT] << dwSA;
}

void RSP_Special_SRL() {		// Shift word Right Logical

	//DWORD dwRS = rs(g_dwRSPOp);	// Should be == 0
	DWORD dwSA = sa(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	g_dwRSPGPR[dwRD] = g_dwRSPGPR[dwRT] >> dwSA;
}
void RSP_Special_SRA() {	// Shift word Right Arithmetic

	//DWORD dwRS = rs(g_dwRSPOp);	// Should be == 0
	DWORD dwSA = sa(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	g_dwRSPGPR[dwRD] = ((s32)g_dwRSPGPR[dwRT]) >> dwSA;
}

void RSP_Special_SLLV() {		// Shift word Left Logical Variable

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	// Sign Extend
	g_dwRSPGPR[dwRD] = g_dwRSPGPR[dwRT] << (g_dwRSPGPR[dwRS]&0x1F);

}
void RSP_Special_SRLV()
{

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	//g_qwGPR[dwRD] = (u64)(((u32)g_qwGPR[dwRT]) >> (g_qwGPR[dwRS]&0x1F));

	g_dwRSPGPR[dwRD] = g_dwRSPGPR[dwRT] >> (g_dwRSPGPR[dwRS]&0x1F);


}
void RSP_Special_SRAV() { WARN_NOIMPL("RSP: SRAV"); }
void RSP_Special_JR() {			// Jump Register

	DWORD dwRS = rs(g_dwRSPOp);

	g_dwNewRSPPC = 0x04000000 | (g_dwRSPGPR[dwRS] & 0x1fff);
	g_nRSPDelay = DO_DELAY;
}

void RSP_Special_JALR()
{
	// Jump And Link
	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);


	g_dwNewRSPPC = 0x04000000 | (g_dwRSPGPR[dwRS] & 0x1fff);
	DBGConsole_Msg(0, "JALR: pc 0x%08x, new: 0x%08x", RSPPC(), g_dwNewRSPPC);
	g_dwRSPGPR[dwRD] = (s32)(RSPPC() + 8);		// Store return address
	
	g_nRSPDelay = DO_DELAY;


}

void RSP_Special_BREAK()
{
	Memory_SP_SetRegisterBits(SP_STATUS_REG, SP_STATUS_SIG2|SP_STATUS_BROKE|SP_STATUS_HALT);

	if(Memory_SP_GetRegister(SP_STATUS_REG) & SP_STATUS_INTR_BREAK)
	{
		Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_SP);
		AddCPUJob(CPU_CHECK_INTERRUPTS);
	}

	DBGConsole_Msg(0, "RSP: [WBREAK]");
	StopRSP();
}

void RSP_Special_ADD() {			// ADD signed - may throw exception

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	// 32 bit numbers for addition...
	g_dwRSPGPR[dwRD] = (s32)g_dwRSPGPR[dwRS] + (s32)g_dwRSPGPR[dwRT];
}
void RSP_Special_ADDU() {			// ADD Unsigned - doesn't throw exception

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	// 32 bit numbers for addition...
	g_dwRSPGPR[dwRD] = (s32)g_dwRSPGPR[dwRS] + (s32)g_dwRSPGPR[dwRT];
}

void RSP_Special_SUB() {			// SUB signed - may throw exception

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	// 32 bit numbers for addition...
	g_dwRSPGPR[dwRD] = (s32)g_dwRSPGPR[dwRS] - (s32)g_dwRSPGPR[dwRT];
}
void RSP_Special_SUBU() {			// SUB Unsigned - may throw exception

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	// 32 bit numbers for addition...
	g_dwRSPGPR[dwRD] = (s32)g_dwRSPGPR[dwRS] - (s32)g_dwRSPGPR[dwRT];
}
void RSP_Special_AND() { 			// logical AND

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);
	g_dwRSPGPR[dwRD] = g_dwRSPGPR[dwRS] & g_dwRSPGPR[dwRT];

}
void RSP_Special_OR() {				// logical OR

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);
	g_dwRSPGPR[dwRD] = g_dwRSPGPR[dwRS] | g_dwRSPGPR[dwRT];

}
void RSP_Special_XOR() {			// logical eXclusive OR


	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);
	g_dwRSPGPR[dwRD] = g_dwRSPGPR[dwRS] ^ g_dwRSPGPR[dwRT];



}
void RSP_Special_NOR() {			// logical NOT OR

	DWORD dwRS = rs(g_dwRSPOp);
	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);
	g_dwRSPGPR[dwRD] = ~(g_dwRSPGPR[dwRS] | g_dwRSPGPR[dwRT]);

}
void RSP_Special_SLT() { WARN_NOIMPL("RSP: SLT"); }
void RSP_Special_SLTU() { WARN_NOIMPL("RSP: SLTU"); }








void RSP_RegImm_Unk(){ WARN_NOEXIST("RSP_RegImm_Unk"); }
void RSP_RegImm_BLTZ(){				// Branch on Less than Zero
	//branch if rs < 0
	DWORD dwRS = rs(g_dwRSPOp);

	if ((s32)g_dwRSPGPR[dwRS] < 0) {
		s16 wOffset = (s16)imm(g_dwRSPOp);
		g_dwNewRSPPC = RSPPC() + (s32)wOffset*4 + 4;
		g_nRSPDelay = DO_DELAY;
	} 
}
void RSP_RegImm_BGEZ(){				// Branch on Greater than or Equal to than Zero
	//branch if rs < 0
	DWORD dwRS = rs(g_dwRSPOp);

	if ((s32)g_dwRSPGPR[dwRS] >= 0) {
		s16 wOffset = (s16)imm(g_dwRSPOp);
		g_dwNewRSPPC = RSPPC() + (s32)wOffset*4 + 4;
		g_nRSPDelay = DO_DELAY;
	}
}
void RSP_RegImm_BLTZAL(){ WARN_NOIMPL("RSP: BLTZAL"); }
void RSP_RegImm_BGEZAL()
{
	//WARN_NOIMPL("RSP: BGEZAL");

	//branch if rs >= 0
	DWORD dwRS = rs(g_dwRSPOp);

	// Apparently this always happens, even if branch not taken
	g_dwRSPGPR[31] = RSPPC() + 8;		// Store return address	
	if ((s32)g_dwRSPGPR[dwRS] >= 0) {
		s16 wOffset = (s16)imm(g_dwRSPOp);
		g_dwNewRSPPC = RSPPC() + ((s32)wOffset*4) + 4;
		g_nRSPDelay = DO_DELAY;

	} 

}











void RSP_Cop0_Unk( ) {WARN_NOEXIST("RSP_Cop0_Unk"); }
void RSP_Cop0_MFC0( ) { 

	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	if(dwRD < 8)	g_dwRSPGPR[dwRT] = Read32Bits(0x84040000 | (dwRD << 2));			// SP Regs
	else			g_dwRSPGPR[dwRT] = Read32Bits(0x84100000 | ((dwRD & ~0x08) << 2));	// DP Command Regs.

}
void RSP_Cop0_MTC0( ) {

	DWORD dwRT = rt(g_dwRSPOp);
	DWORD dwRD = rd(g_dwRSPOp);

	if(dwRD < 8) Write32Bits(0x84040000 | (dwRD << 2), g_dwRSPGPR[dwRT]);			// SP Regs
	else		 Write32Bits(0x84100000 | ((dwRD & ~0x08) << 2),g_dwRSPGPR[dwRT]);	// DP Command Regs.
}


void RSP_LWC2_Unk() { WARN_NOEXIST("RSP_LWC2_Unk"); }
void RSP_LWC2_LBV() { WARN_NOIMPL("RSP: LBV"); }
void RSP_LWC2_LLV() { WARN_NOIMPL("RSP: LLV"); }
void RSP_LWC2_LRV() { WARN_NOIMPL("RSP: LRV"); }
void RSP_LWC2_LPV() { WARN_NOIMPL("RSP: LPV"); }
void RSP_LWC2_LUV() { WARN_NOIMPL("RSP: LUV"); }
void RSP_LWC2_LHV() { WARN_NOIMPL("RSP: LHV"); }
void RSP_LWC2_LFV() { WARN_NOIMPL("RSP: LFV"); }
void RSP_LWC2_LWV() { WARN_NOIMPL("RSP: LWV"); }
void RSP_LWC2_LTV() { WARN_NOIMPL("RSP: LTV"); }

/*
  +-----------+---------------------------------------------------+
  | LxV       | Load xxxxxx to vector                             |
  +-----------+---------+---------+---------+-------+-+-----------+
  |  110010   |  base   |  dest   |  xxxxx  |  sel  |0|   offset  |
  +-----6-----+----5----+----5----+----5----+---4---+1+-----6-----+
  Format:  LxV $v<dest>[sel], offset(base)
*/

void RSP_LWC2_LSV() {

	DWORD dwBase = base(g_dwRSPOp);
	DWORD dwDest  = dest(g_dwRSPOp);
	DWORD dwV = sel(g_dwRSPOp);
	s16 wOffset = rsp_vecoffset(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + ((s32)wOffset<<1)) & 0x0FFF;
	BYTE *pBase = (BYTE *)MemoryBuffers[MEM_SP_MEM];

	// Divide by two - it points to 8bits in the registers, not 16bits
	dwV >>= 1;
    g_wVR[dwDest][dwV+0] = (pBase[(dwAddress+0)^3] << 8) | pBase[(dwAddress+1)^3];

}


/* The following will read from 0(s0) and 0(s0), 8(s0) and 8(s0) respectively
0x04001514: <ca191800> LDV       $vec25[0h] <- 0x0000(s0)
0x04001518: <ca181c00> LDV       $vec24[4h] <- 0x0000(s0)
0x0400151c: <ca171801> LDV       $vec23[0h] <- 0x0008(s0)
0x04001520: <ca171c01> LDV       $vec23[4h] <- 0x0008(s0)

*/
void RSP_LWC2_LDV() {			// Load Doubleword into Vector

	DWORD dwBase = base(g_dwRSPOp);
	DWORD dwDest  = dest(g_dwRSPOp);
	DWORD dwV = sel(g_dwRSPOp);
	s16 wOffset = rsp_vecoffset(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + ((s32)wOffset<<3)) & 0x0FFF;
	BYTE *pBase = (BYTE *)MemoryBuffers[MEM_SP_MEM];

	// Check for dwV>>1+x being > 7 here?
	dwV >>= 1;
    g_wVR[dwDest][dwV+0] = (pBase[(dwAddress+0)^3] << 8) | pBase[(dwAddress+1)^3];
    g_wVR[dwDest][dwV+1] = (pBase[(dwAddress+2)^3] << 8) | pBase[(dwAddress+3)^3];
    g_wVR[dwDest][dwV+2] = (pBase[(dwAddress+4)^3] << 8) | pBase[(dwAddress+5)^3];
    g_wVR[dwDest][dwV+3] = (pBase[(dwAddress+6)^3] << 8) | pBase[(dwAddress+7)^3];
}

void RSP_LWC2_LQV() { 


	DWORD dwBase	= base(g_dwRSPOp);
	DWORD dwDest	= dest(g_dwRSPOp);
	DWORD dwV		= sel(g_dwRSPOp);
	s16 wOffset = rsp_vecoffset(g_dwRSPOp);

	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + ((s32)wOffset<<4)) & 0x0FFF;
	BYTE *pBase = (BYTE *)MemoryBuffers[MEM_SP_MEM];

	// This is shit. It should load into the whole vector (i.e 128bits)

	for(DWORD i=0; i < (16 - (dwAddress & 0xf)); i++) {

		if((dwV+i) & 1) {
				g_wVR[dwDest][((dwV+i)>>1)] &= 0xff00;
				g_wVR[dwDest][((dwV+i)>>1)] |= pBase[(dwAddress+i)^3];
		} else {
				g_wVR[dwDest][((dwV+i)>>1)] &= 0x00ff;
				g_wVR[dwDest][((dwV+i)>>1)] |= pBase[(dwAddress+i)^3] << 8;
		}
	}
}


void RSP_SWC2_Unk() {  WARN_NOEXIST("RSP_SWC2_Unk"); }
void RSP_SWC2_SBV() { WARN_NOIMPL("RSP: SBV"); }
void RSP_SWC2_SSV() {		// Store single to vector (16bits)
	
	DWORD dwBase = base(g_dwRSPOp);
	DWORD dwDest  = dest(g_dwRSPOp);
	DWORD dwV = sel(g_dwRSPOp);
	s16 wOffset = rsp_vecoffset(g_dwRSPOp);


	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + ((s32)wOffset<<1)) & 0x0FFF;
	BYTE *pBase = (BYTE *)MemoryBuffers[MEM_SP_MEM];

	dwV >>= 1;
    pBase[(dwAddress+0)^3] = (BYTE)(g_wVR[dwDest][dwV+0] >> 8);
	pBase[(dwAddress+1)^3] = (BYTE)(g_wVR[dwDest][dwV+0] & 0xFF);
}

void RSP_SWC2_SLV() {		// Store word to vector (32bits)

	DWORD dwBase = base(g_dwRSPOp);
	DWORD dwDest  = dest(g_dwRSPOp);
	DWORD dwV = sel(g_dwRSPOp);
	s16 wOffset = rsp_vecoffset(g_dwRSPOp);


	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + ((s32)wOffset<<2)) & 0x0FFF;
	BYTE *pBase = (BYTE *)MemoryBuffers[MEM_SP_MEM];

	dwV >>= 1;
	pBase[(dwAddress+0)^3] = (BYTE)(g_wVR[dwDest][dwV+0] >> 8);
	pBase[(dwAddress+1)^3] = (BYTE)(g_wVR[dwDest][dwV+0]);
	pBase[(dwAddress+2)^3] = (BYTE)(g_wVR[dwDest][dwV+1] >> 8);
	pBase[(dwAddress+3)^3] = (BYTE)(g_wVR[dwDest][dwV+1]);

}
void RSP_SWC2_SDV() {			// Store doubleword from vector (64 bits)

	DWORD dwBase = base(g_dwRSPOp);
	DWORD dwDest  = dest(g_dwRSPOp);
	DWORD dwV = sel(g_dwRSPOp);
	s16 wOffset = rsp_vecoffset(g_dwRSPOp);


	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + ((s32)wOffset<<3)) & 0x0FFF;
	BYTE *pBase = (BYTE *)MemoryBuffers[MEM_SP_MEM];

	// Check for dwV>>1+x being > 7 here?
	dwV >>= 1;
	pBase[(dwAddress+0)^3] = (BYTE)(g_wVR[dwDest][dwV+0] >> 8);
	pBase[(dwAddress+1)^3] = (BYTE)(g_wVR[dwDest][dwV+0]);
	pBase[(dwAddress+2)^3] = (BYTE)(g_wVR[dwDest][dwV+1] >> 8);
	pBase[(dwAddress+3)^3] = (BYTE)(g_wVR[dwDest][dwV+1]);
	pBase[(dwAddress+4)^3] = (BYTE)(g_wVR[dwDest][dwV+2] >> 8);
	pBase[(dwAddress+5)^3] = (BYTE)(g_wVR[dwDest][dwV+2]);
	pBase[(dwAddress+6)^3] = (BYTE)(g_wVR[dwDest][dwV+3] >> 8);
	pBase[(dwAddress+7)^3] = (BYTE)(g_wVR[dwDest][dwV+3]);
}

void RSP_SWC2_SQV() {

	DWORD dwBase = base(g_dwRSPOp);
	DWORD dwDest = dest(g_dwRSPOp);
	DWORD dwV	 = sel(g_dwRSPOp);
	s16 wOffset = rsp_vecoffset(g_dwRSPOp);


	DWORD dwAddress = (u32)((s32)g_dwRSPGPR[dwBase] + ((s32)wOffset<<4)) & 0x0FFF;
	BYTE *pBase = (BYTE *)MemoryBuffers[MEM_SP_MEM];

	for(DWORD i=0; i < (16 - (dwAddress & 0xf)); i++) {

		if((dwV+i) & 1) {
			pBase[(dwAddress+i)^3] = (BYTE)(g_wVR[dwDest][((dwV+i)>>1)]);
		} else {
			pBase[(dwAddress+i)^3] = (BYTE)(g_wVR[dwDest][((dwV+i)>>1)] >> 8);
		}
	}
}

void RSP_SWC2_SRV() { WARN_NOIMPL("RSP: SRV"); }
void RSP_SWC2_SPV() { WARN_NOIMPL("RSP: SPV"); }
void RSP_SWC2_SUV() { WARN_NOIMPL("RSP: SUV"); }
void RSP_SWC2_SHV() { WARN_NOIMPL("RSP: SHV"); }
void RSP_SWC2_SFV() { WARN_NOIMPL("RSP: SFV"); }
void RSP_SWC2_SWV() { WARN_NOIMPL("RSP: SWV"); }
void RSP_SWC2_STV() { WARN_NOIMPL("RSP: STV"); }


void RSP_Cop2_Unk()	{ WARN_NOEXIST("RSP_Cop2_Unk"); }
void RSP_Cop2_Vector()	{ RSPCop2VInstruction[funct(g_dwRSPOp)](); }
void RSP_Cop2_MFC2()	{ WARN_NOIMPL("RSP: MFC2"); }
void RSP_Cop2_CFC2()	{ WARN_NOIMPL("RSP: CFC2"); }
void RSP_Cop2_MTC2()	{ WARN_NOIMPL("RSP: MTC2"); }
void RSP_Cop2_CTC2()	{ WARN_NOIMPL("RSP: CTC2"); }

void RSP_Cop2V_Unk()		{ WARN_NOEXIST("RSP_Cop2V_Unk"); }
void RSP_Cop2V_VMULF()	{ /*WARN_NOIMPL("RSP: VMULF");*/ }
void RSP_Cop2V_VMULU()	{ WARN_NOIMPL("RSP: VMULU"); }
void RSP_Cop2V_VRNDP()	{ WARN_NOIMPL("RSP: VRNDP"); }
void RSP_Cop2V_VMULQ()	{ WARN_NOIMPL("RSP: VMULQ"); }
void RSP_Cop2V_VMUDL()	{ /*WARN_NOIMPL("RSP: VMUDL");*/ }
void RSP_Cop2V_VMUDM()	{ /*WARN_NOIMPL("RSP: VMUDM");*/ }
void RSP_Cop2V_VMUDN()	{ /*WARN_NOIMPL("RSP: VMUDN");*/ }
void RSP_Cop2V_VMUDH()	{ /*WARN_NOIMPL("RSP: VMUDH");*/ }
void RSP_Cop2V_VMACF()	{ /*WARN_NOIMPL("RSP: VMACF");*/ }
void RSP_Cop2V_VMACU()	{ WARN_NOIMPL("RSP: VMACU"); }
void RSP_Cop2V_VRNDN()	{ WARN_NOIMPL("RSP: VRNDN"); }
void RSP_Cop2V_VMACQ()	{ WARN_NOIMPL("RSP: VMACQ"); }
void RSP_Cop2V_VMADL()	{ WARN_NOIMPL("RSP: VMADL"); }
void RSP_Cop2V_VMADM()	{ /*WARN_NOIMPL("RSP: VMADM");*/ }
void RSP_Cop2V_VMADN()	{ /*WARN_NOIMPL("RSP: VMADN");*/ }
void RSP_Cop2V_VMADH()	{ /*WARN_NOIMPL("RSP: VMADH");*/ }

void RSP_Cop2V_VADD()	{

	DWORD dwDest	= (g_dwRSPOp >>  6) & 0x1f;
	DWORD dwA		= (g_dwRSPOp >> 11) & 0x1f;
	DWORD dwB		= (g_dwRSPOp >> 16) & 0x1f;
	DWORD dwPattern = (g_dwRSPOp >> 21) & 0x0f;
	LOAD_VECTORS(dwA, dwB, dwPattern);
	for (DWORD dwV = 0; dwV < 8; dwV++) {
		g_wVR[dwDest][dwV] = g_wTemp[0][dwV] + g_wTemp[1][dwV];

		// Add 1 if carry flag set
		if (g_dwCarryFlags & (1<<dwV))
			g_wVR[dwDest][dwV]++;

		// Clip?
        if((~g_wTemp[0][dwV] & ~g_wTemp[1][dwV] & g_wVR[dwDest][dwV]) & 0x8000)
                g_wVR[dwDest][dwV] = 0x7fff;
        else if((g_wTemp[0][dwV] & g_wTemp[1][dwV] & ~g_wVR[dwDest][dwV]) & 0x8000)
                g_wVR[dwDest][dwV] = 0x8000;
	}


	// To set the carry flags
	/*g_dwCarryFlags = 0;
	for (dwV = 0; dwV < 8; dwV++) {
		if(((s32)g_wTemp[0][dwV] + (s32)g_wTemp[1][dwV]) != (s32)g_wVR[dwDest][dwV])
			g_dwCarryFlags |= (1 << dwV);	// Set carry flag
	}*/
}

void RSP_Cop2V_VSUB()	{ /*WARN_NOIMPL("RSP: VSUB");*/ }
void RSP_Cop2V_VSUT()	{ WARN_NOIMPL("RSP: VSUT"); }
void RSP_Cop2V_VABS()	{ WARN_NOIMPL("RSP: VABS"); }
void RSP_Cop2V_VADDC()	{ /*WARN_NOIMPL("RSP: VADDC");*/ }
void RSP_Cop2V_VSUBC()	{ /*WARN_NOIMPL("RSP: VSUBC");*/ }
void RSP_Cop2V_VADDB()	{ WARN_NOIMPL("RSP: VADDB"); }
void RSP_Cop2V_VSUBB()	{ WARN_NOIMPL("RSP: VACCB"); }
void RSP_Cop2V_VACCB()	{ WARN_NOIMPL("RSP: VACCB"); }
void RSP_Cop2V_VSUCB()	{ WARN_NOIMPL("RSP: VSUCB"); }
void RSP_Cop2V_VSAD()	{ WARN_NOIMPL("RSP: VSAD"); }
void RSP_Cop2V_VSAC()	{ WARN_NOIMPL("RSP: VSAC"); }
void RSP_Cop2V_VSUM()	{ WARN_NOIMPL("RSP: VSUM"); }
void RSP_Cop2V_VSAW()	{ /*WARN_NOIMPL("RSP: VSAW");*/ }
void RSP_Cop2V_VLT()		{ WARN_NOIMPL("RSP: VLT"); }
void RSP_Cop2V_VEQ()		{ WARN_NOIMPL("RSP: VEQ"); }
void RSP_Cop2V_VNE()		{ WARN_NOIMPL("RSP: VNE"); }
void RSP_Cop2V_VGE()		{ WARN_NOIMPL("RSP: VGE"); }
void RSP_Cop2V_VCL()		{ /*WARN_NOIMPL("RSP: VCL");*/ }
void RSP_Cop2V_VCH()		{ WARN_NOIMPL("RSP: VCH"); }
void RSP_Cop2V_VCR()		{ WARN_NOIMPL("RSP: VCR"); }
void RSP_Cop2V_VMRG()	{ WARN_NOIMPL("RSP: VMRG"); }
void RSP_Cop2V_VAND()	{
	
	DWORD dwDest	= (g_dwRSPOp >>  6) & 0x1f;
	DWORD dwA		= (g_dwRSPOp >> 11) & 0x1f;
	DWORD dwB		= (g_dwRSPOp >> 16) & 0x1f;
	DWORD dwPattern = (g_dwRSPOp >> 21) & 0x0f;
	LOAD_VECTORS(dwA, dwB, dwPattern);

	for (DWORD dwV = 0; dwV < 8; dwV++) {
		//DPF(DEBUG_ALWAYS, "%d: A: 0x%04x B: 0x%04x", dwV, g_wTemp[0][dwV], g_wTemp[1][dwV]);
		g_wVR[dwDest][dwV] = g_wTemp[0][dwV] & g_wTemp[1][dwV];
		//DPF(DEBUG_ALWAYS, "%d: D: 0x%04x", dwV, g_wVR[dwDest][dwV]);
	}
}
void RSP_Cop2V_VNAND()	{
	DWORD dwDest	= (g_dwRSPOp >>  6) & 0x1f;
	DWORD dwA		= (g_dwRSPOp >> 11) & 0x1f;
	DWORD dwB		= (g_dwRSPOp >> 16) & 0x1f;
	DWORD dwPattern = (g_dwRSPOp >> 21) & 0x0f;
	LOAD_VECTORS(dwA, dwB, dwPattern);
	for (DWORD dwV = 0; dwV < 8; dwV++) {
		g_wVR[dwDest][dwV] = ~(g_wTemp[0][dwV] & g_wTemp[1][dwV]);
	}
}
void RSP_Cop2V_VOR()		{

	DWORD dwDest	= (g_dwRSPOp >>  6) & 0x1f;
	DWORD dwA		= (g_dwRSPOp >> 11) & 0x1f;
	DWORD dwB		= (g_dwRSPOp >> 16) & 0x1f;
	DWORD dwPattern = (g_dwRSPOp >> 21) & 0x0f;
	LOAD_VECTORS(dwA, dwB, dwPattern);
	for (DWORD dwV = 0; dwV < 8; dwV++) {
		g_wVR[dwDest][dwV] = g_wTemp[0][dwV] | g_wTemp[1][dwV];
	}

}
void RSP_Cop2V_VNOR()	{

	DWORD dwDest	= (g_dwRSPOp >>  6) & 0x1f;
	DWORD dwA		= (g_dwRSPOp >> 11) & 0x1f;
	DWORD dwB		= (g_dwRSPOp >> 16) & 0x1f;
	DWORD dwPattern = (g_dwRSPOp >> 21) & 0x0f;
	LOAD_VECTORS(dwA, dwB, dwPattern);
	for (DWORD dwV = 0; dwV < 8; dwV++) {
		g_wVR[dwDest][dwV] = ~(g_wTemp[0][dwV] | g_wTemp[1][dwV]);
	}
}

void RSP_Cop2V_VXOR()	{

	DWORD dwDest	= (g_dwRSPOp >>  6) & 0x1f;
	DWORD dwA		= (g_dwRSPOp >> 11) & 0x1f;
	DWORD dwB		= (g_dwRSPOp >> 16) & 0x1f;
	DWORD dwPattern = (g_dwRSPOp >> 21) & 0x0f;
	LOAD_VECTORS(dwA, dwB, dwPattern);
	for (DWORD dwV = 0; dwV < 8; dwV++) {
		g_wVR[dwDest][dwV] = g_wTemp[0][dwV] ^ g_wTemp[1][dwV];
	}

}
void RSP_Cop2V_VNXOR()	{

	DWORD dwDest	= (g_dwRSPOp >>  6) & 0x1f;
	DWORD dwA		= (g_dwRSPOp >> 11) & 0x1f;
	DWORD dwB		= (g_dwRSPOp >> 16) & 0x1f;
	DWORD dwPattern = (g_dwRSPOp >> 21) & 0x0f;
	LOAD_VECTORS(dwA, dwB, dwPattern);
	for (DWORD dwV = 0; dwV < 8; dwV++) {
		g_wVR[dwDest][dwV] = ~(g_wTemp[0][dwV] ^ g_wTemp[1][dwV]);
	}
}

void RSP_Cop2V_VRCP()	{ WARN_NOIMPL("RSP: VRCP"); }
void RSP_Cop2V_VRCPL()	{ WARN_NOIMPL("RSP: VRCPL"); }
void RSP_Cop2V_VRCPH()	{ WARN_NOIMPL("RSP: VRCPH"); }
void RSP_Cop2V_VMOV()	{ WARN_NOIMPL("RSP: VMOV"); }
void RSP_Cop2V_VRSQ()	{ WARN_NOIMPL("RSP: VRSQ"); }
void RSP_Cop2V_VRSQL()	{ WARN_NOIMPL("RSP: VRSQL"); }
void RSP_Cop2V_VRSQH()	{ WARN_NOIMPL("RSP: VRSQH"); }