/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __ASM_H__
#define __ASM_H__

// Get the lo and high half indices for MIPS registers
#define lohalf(x)		(x*2 + 0)
#define hihalf(x)		(x*2 + 1)

// Emulated regs 0..31 here

// Intel register codes. Odd ordering is for intel bytecode
enum REGCODE {
	EAX_CODE = 0,
	ECX_CODE = 1,
	EDX_CODE = 2,
	EBX_CODE = 3,
	ESP_CODE = 4,
	EBP_CODE = 5,
	ESI_CODE = 6,
	EDI_CODE = 7
};

// Some more friendly macros:
#define SetVar32(VarName, Value) \
	EmitWORD(pCode, 0x05c7); EmitDWORD(pCode, (DWORD)&VarName); EmitDWORD(pCode, Value);



#define CRASH(x)			EmitWORD(pCode, 0x05c7); EmitDWORD(pCode, 0x00000000); EmitDWORD(pCode, x) //c7 05 xx+0 xx xx xx ef be ad de

#define CALL_EAX()			EmitWORD(pCode, 0xd0ff)
#define RET()				EmitBYTE(pCode, 0xC3)

#define PUSH(reg)			EmitBYTE(pCode, 0x50 | reg)
#define POP(reg)			EmitBYTE(pCode, 0x58 | reg)
		
#define SAVE_ESP() MOV_MEM_REG(&g_SavedESP, ESP_CODE)
#define SAVE_EBP() MOV_MEM_REG(&g_SavedEBP, EBP_CODE)
#define RESTORE_ESP() if (g_bESPInvalid) { MOV_REG_MEM(ESP_CODE, &g_SavedESP); g_bESPInvalid = FALSE; }
#define RESTORE_EBP() if (g_bEBPInvalid) { MOV_REG_MEM(EBP_CODE, &g_SavedEBP); g_bEBPInvalid = FALSE; }

					// e.g ADD_EAX_EBX = 0xc303
#define ADD(reg1, reg2)		EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x03)
#define ADC(reg1, reg2)		EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x13)
#define SUB(reg1, reg2)		EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x2b)
#define SBB(reg1, reg2)		EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x1b)

#define AND(reg1, reg2)		if (reg1 != reg2) {EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x23);}
#define MOV(reg1, reg2)		if (reg1 != reg2) {EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x8B);}
#define  OR(reg1, reg2)		if (reg1 != reg2) {EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x0B);}
#define XCHG(reg1,reg2)		if (reg1 != reg2) {EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x87);}
#define XOR(reg1, reg2)		EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x33)
#define CMP(reg1, reg2)     EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x3b)
#define TEST(reg1, reg2)    EmitWORD(pCode, ((0xc0 | (reg1<<3) | reg2)<<8) | 0x85)

#define MOVSX(reg1, reg2)		EmitWORD(pCode, 0xbf0f); EmitBYTE(pCode, (0xc0 | (reg1<<3) | reg2))

#define SHLI(reg, sa)		EmitWORD(pCode, ((0xe0 | reg)<<8) | 0xc1); EmitBYTE(pCode, (BYTE)(sa))
#define SHRI(reg, sa)		EmitWORD(pCode, ((0xe8 | reg)<<8) | 0xc1); EmitBYTE(pCode, (BYTE)(sa))
#define SARI(reg, sa)		EmitWORD(pCode, ((0xf8 | reg)<<8) | 0xc1); EmitBYTE(pCode, (BYTE)(sa))

#define ADDI(reg, data) /*if (reg == EAX_CODE){ EmitBYTE(pCode, 0x05); } else */{ EmitWORD(pCode, ((0xc0 | reg)<<8) | 0x81);}	\
						EmitDWORD(pCode, data)

// Use short form (0x83d0) if data is just one byte!
#define ADCI(reg, data) \
	/*if (reg == EAX_CODE){ EmitBYTE(pCode, 0x15); EmitDWORD(pCode, data)} else*/ \
	if (data <= 127 && data > -127)							\
	{ 														\
		EmitWORD(pCode, ((0xd0 | reg)<<8) | 0x83);			\
		EmitBYTE(pCode, (BYTE)(data));						\
	}														\
	else													\
	{														\
		EmitWORD(pCode, ((0xd0 | reg)<<8) | 0x81);			\
		EmitDWORD(pCode, data);								\
	}


#define CMPI(reg, data) EmitWORD(pCode, ((0xf8 | reg)<<8) | 0x81); EmitDWORD(pCode, data)

#define MUL(reg) EmitWORD(pCode, ((0xe0 | reg)<<8) | 0xf7)
#define MUL_EAX_MEM(mem) EmitBYTE(pCode, 0xf7); EmitBYTE(pCode, 0x25); EmitDWORD(pCode, (DWORD)(mem));


#define ADDI_b(reg, data) EmitWORD(pCode, ((0xc0 | reg)<<8) | 0x83); EmitBYTE(pCode, data)
#define ADCI_b(reg, data) EmitWORD(pCode, ((0xd0 | reg)<<8) | 0x83); EmitBYTE(pCode, data)

#define DEC(reg) EmitBYTE(pCode, 0x48 | (reg))



#define MOV_REG_MEM(reg, mem) /*if (reg == EAX_CODE) { EmitBYTE(pCode, 0xa1); } else*/ { EmitWORD(pCode,  (((reg<<3) | 0x05)<<8) | 0x8b);} \
						      EmitDWORD(pCode, (DWORD)(mem))


#define MOV_MEM_REG(mem, reg) /*if (reg == EAX_CODE) { EmitBYTE(pCode, 0xa3); } else*/ { EmitWORD(pCode,  (((reg<<3) | 0x05)<<8) | 0x89);} \
							  EmitDWORD(pCode, (DWORD)(mem))


#define MOVI(reg, data)		EmitBYTE(pCode, 0xB8 | (reg)); EmitDWORD(pCode, (data))	


#define MOVI_MEM(mem, data) EmitWORD(pCode, 0x05C7); EmitDWORD(pCode, (DWORD)(mem)); EmitDWORD(pCode, data)

#define ANDI(reg, data) /*if (reg == EAX_CODE) { EmitBYTE(pCode, 0x25); } else */{ EmitWORD(pCode, ((0xe0 | reg)<<8) | 0x81); } \
						EmitDWORD(pCode, (DWORD)(data))

#define XORI(reg, data) /*if (reg == EAX_CODE) { EmitBYTE(pCode, 0x35); } else */{ EmitWORD(pCode, ((0xf0 | reg)<<8) | 0x81); } \
						EmitDWORD(pCode, (DWORD)(data))

#define ORI(reg, data) /*if (reg == EAX_CODE) { EmitBYTE(pCode, 0x0D); } else*/ { EmitWORD(pCode, ((0xc8 | reg)<<8) | 0x81); } \
					   EmitDWORD(pCode, (DWORD)(data))


#define SETNS(reg)	EmitWORD(pCode, 0x990f); EmitBYTE(pCode, 0xc0 | (reg))
#define SETL(reg)	EmitWORD(pCode, 0x9c0f); EmitBYTE(pCode, 0xc0 | (reg))
#define SETB(reg)	EmitWORD(pCode, 0x920f); EmitBYTE(pCode, 0xc0 | (reg))
#define JA(off)		EmitWORD(pCode, ((off)<<8) | 0x77)
#define JB(off)		EmitWORD(pCode, ((off)<<8) | 0x72)
#define JAE(off)	EmitWORD(pCode, ((off)<<8) | 0x73)
#define JE(off)		EmitWORD(pCode, ((off)<<8) | 0x74)
#define JNE(off)	EmitWORD(pCode, ((off)<<8) | 0x75)
#define JL(off)		EmitWORD(pCode, ((off)<<8) | 0x7C)
#define JGE(off)	EmitWORD(pCode, ((off)<<8) | 0x7D)
#define JLE(off)	EmitWORD(pCode, ((off)<<8) | 0x7E)
#define JG(off)		EmitWORD(pCode, ((off)<<8) | 0x7F)
#define JMP(off)	EmitWORD(pCode, ((off)<<8) | 0xeb)
#define JMP_DIRECT(off)	EmitBYTE(pCode, 0xe9); EmitDWORD(pCode, (DWORD)(off))

#define CDQ			EmitBYTE(pCode, 0x99)		//cdq						// 99	
		

#define AND_EAX(data) EmitBYTE(pCode, 0x25);   EmitDWORD(pCode, (DWORD)(data))	//and		eax, 0x12345678 // 25 xx xx xx xx
#define AND_EBX(data) EmitWORD(pCode, 0xe381); EmitDWORD(pCode, (DWORD)(data))	//and		ebx, 0x12345678 // 81 e3 xx xx xx xx


#define PXOR(mm)	EmitWORD(pCode, 0xed0f); EmitBYTE(pCode, 0xc0 + (mm)*9);
#define EMMS		EmitWORD(pCode, 0x770f);


#define FLD_MEM(mem)  { EmitWORD(pCode, 0x05d9);} \
						      EmitDWORD(pCode, (DWORD)(mem))

#define FADD_MEM(mem) { EmitWORD(pCode, 0x05d8);} \
						      EmitDWORD(pCode, (DWORD)(mem))
#define FSUB_MEM(mem) { EmitWORD(pCode, 0x25d8);} \
						      EmitDWORD(pCode, (DWORD)(mem))
#define FMUL_MEM(mem) { EmitWORD(pCode, 0x0dd8);} \
						      EmitDWORD(pCode, (DWORD)(mem))
#define FDIV_MEM(mem) { EmitWORD(pCode, 0x35d8);} \
						      EmitDWORD(pCode, (DWORD)(mem))

#define FSTP_MEM(mem) { EmitWORD(pCode,  0x1dd9);} \
						      EmitDWORD(pCode, (DWORD)(mem))

#define FSQRT EmitWORD(pCode, 0xfad9);
#define FCHS EmitWORD(pCode, 0xe0d9);

//////////////////////////////////////
#define FLD_MEMp(pCode, mem)  { EmitWORD(pCode, 0x05d9);} \
						      EmitDWORD(pCode, (DWORD)(mem))

#define FSTP_MEMp(pCode, mem) { EmitWORD(pCode,  0x1dd9);} \
						      EmitDWORD(pCode, (DWORD)(mem))

#define FLD(pCode, i)	EmitWORD(pCode, (WORD)((0xc0 + (i))<<8) |(0xd9)); 


#define FXCH(pCode, i)	EmitWORD(pCode, (WORD)((0xc8 + (i))<<8) |(0xd9)); 

#define FFREE(pCode, i)	EmitWORD(pCode, (WORD)((0xc0 + (i))<<8) |(0xdd)); 
#define FINCSTP(pCode) EmitWORD(pCode, 0xf7d9);
#define FSQRTp(pCode) EmitWORD(pCode, 0xfad9);
#define FCHSp(pCode) EmitWORD(pCode, 0xe0d9);

// fadd st(0),st(i)
#define FADD(pCode, i)	EmitWORD(pCode, (WORD)((0xc0 + (i))<<8) |(0xd8)); 

// fsub st(0),st(i)
#define FSUB(pCode, i)	EmitWORD(pCode, (WORD)((0xe0 + (i))<<8) |(0xd8)); 

// fmul st(0),st(i)
#define FMUL(pCode, i)	EmitWORD(pCode, (WORD)((0xc8 + (i))<<8) |(0xd8)); 

// fdiv st(0),st(i)
#define FDIV(pCode, i)	EmitWORD(pCode, (WORD)((0xf0 + (i))<<8) |(0xd8)); 





#endif